// *****************************************************************************************************
// BrianHG_GFX_Layer_mixer.sv
// Takes in multiple layers of 32bit RGBA video, Parallel and serial:
// Offers a 24bit RGB background color for below bottom layer, IE all pixels on all active layers happen to be transparent.
// Offers adjustment offset to the Alpha channel's opacity of each layer.
// Then offers serial layer order swapping of each parallel layer channel.
// Then offers parallel layer order swapping.
//
// Then mixes all parallel streams.
//
// Then mixes all resulting serial layers into the 1 final output stream.
//
// Version 1.6, January 2, 2022.
//
// Written by Brian Guralnick.
// For public use.
//
//
// See: https://www.eevblog.com/forum/fpga/brianhg_ddr3_controller-open-source-ddr3-controller/
//
// *****************************************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// Typedef structures.
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
typedef struct packed {              // Generate a structure for the sync generator bus.
                logic  [2:0] phase ;
                logic        h_ena ;
                logic        v_ena ;
                logic        vena  ;
                logic        hs    ;
                logic        vs    ;
                logic        wlena ;
                logic [31:0] rgba  ;
                } vlb_bus ;


module BrianHG_GFX_Layer_mixer #(

parameter bit [3:0]  PDI_LAYERS             = 1,                       // Number of parallel layered 'BrianHG_GFX_Video_Line_Buffer' modules in the system.  1 through 8 is allowed.
parameter bit [3:0]  SDI_LAYERS             = 1,                       // Use 1,2,4, or 8 sequential display layers in each 'BrianHG_GFX_Video_Line_Buffer' module in the system.

parameter bit        ENABLE_alpha_adj       = 1,                       // Use 0 to bypass the alpha adjustment logic
parameter bit        ENABLE_SDI_layer_swap  = 1,                       // Use 0 to bypass the serial layer swapping logic
parameter bit        ENABLE_PDI_layer_swap  = 1,                       // Use 0 to bypass the parallel layer swapping logic

// ******* Do not edit these ****
parameter bit [6:0]  LAYERS                 = PDI_LAYERS * SDI_LAYERS  // Total window layers in system
)(

// *******************************************************************************
// ***** Layer alpha override and order re-positioning control inputs. 
// *******************************************************************************
input        [2:0]         CLK_DIVIDER                        , // Supports 0 through 7 to divide the clock from 1 through 8.
                                                                // Also cannot be higher than SDI_LAYERS and only SDI layers 0 through this number will be shown.

input       [23:0]         CMD_BGC_RGB                        , // Bottom background color when every layer's pixel happens to be transparent. 

input        [2:0]         CMD_SDI_layer_swap [0:PDI_LAYERS-1], // Re-position the SDI layer order of each PDI layer line buffer's output stream. (A Horizontal SDI PHASE layer swap / PDI layer)
input        [2:0]         CMD_PDI_layer_swap [0:SDI_LAYERS-1], // Re-position the PDI layer order of each SDI layer. (A PDI Vertical swap/SDI Layer PHASE) 

// *********************************************************************************************
// **** Video clock domain and input timing from at least 1 BrianHG_GFX_Video_Line_Buffer.
// *********************************************************************************************
input                      VID_RST                            , // Video output pixel clock's reset.
input                      VID_CLK                            , // Video output pixel clock.

input        [2:0]         VCLK_PHASE_IN                      , // Used with sync gen is there are 
input                      H_ena_in                           , // Horizontal video enable.
input                      V_ena_in                           , // Vertical video enable.
input                      VENA_in                            , // High during active video.
input                      HS_in                              , // Horizontal sync output.
input                      VS_in                              , // Vertical sync output.

// ***************************************************************************************
// **** Video video picture stream from every BrianHG_GFX_Video_Line_Buffer PDI layer
// **** with the embedded SDI layers on each line buffer PDI layer output.
// ***************************************************************************************
input                      WLENA_in           [0:PDI_LAYERS-1], // Window Layer Active In.
input        [31:0]        RGBA_in            [0:PDI_LAYERS-1], // 32 bit Video picture data input: Reg, Green, Blue, Alpha-Blend
input        [7:0]         alpha_adj_in       [0:PDI_LAYERS-1], // When 0, the layer translucency will be determined by the graphic data.
                                                                // Any figure from +1 to +127 will progressive force all the graphics opaque.
                                                                // Any figure from -1 to -128 will progressive force all the graphics transparent.

// ***************************************************************************************
// **** Singular mixed video output channel.
// ***************************************************************************************
output       [31:0]        RGBA_out                           , // 32 bit Video picture data output: Reg, Green, Blue, Alpha-Blend

output       [2:0]         VCLK_PHASE_OUT                     , // Pixel clock divider position.
output                     H_ena_out                          , // Horizontal video enable.
output                     V_ena_out                          , // Vertical video enable.
output                     VENA_out                           , // High during active video.
output                     HS_out                             , // Horizontal sync output.
output                     VS_out                               // Vertical sync output.
);


generate
if ( (SDI_LAYERS!=1) && (SDI_LAYERS!=2) && (SDI_LAYERS!=4) && (SDI_LAYERS!=8) )  initial begin
$warning("**************************************");
$warning("*** BrianHG_GFX_Layer_mixer ERROR. ***");
$warning("***********************************************************");
$warning("*** Your current parameter .SDI_LAYERS(%d) is invalid. ***",6'(SDI_LAYERS));
$warning("*** It can only be 1, 2, 4, or 8.                       ***");
$warning("***********************************************************");
$error;
$stop;
end
if ( (PDI_LAYERS<1) || (PDI_LAYERS>8) )  initial begin
$warning("***************************************");
$warning("*** BrianHG_GFX_Layer_mixer ERROR.  ***");
$warning("***********************************************************");
$warning("*** Your current parameter .PDI_LAYERS(%d) is invalid. ***",6'(PDI_LAYERS));
$warning("*** It can only be anywhere from 1 through 8.           ***");
$warning("***********************************************************");
$error;
$stop;
end
endgenerate


vlb_bus  vlb_vid_in  [0:PDI_LAYERS-1] ;
vlb_bus  alpha_out   [0:PDI_LAYERS-1] ;
vlb_bus  sdi_swap    [0:PDI_LAYERS-1] ;
vlb_bus  pdi_swap    [0:PDI_LAYERS-1] ;
vlb_bus  pdi_mix_out [0:0]            ;
vlb_bus  pdi_mix                      ;
vlb_bus  pdi_mix_dly                  ;
vlb_bus  sdi_mix                      ;

 // PDI layer by PDI layer Copy input bus to vlb structure.
always_comb for (int i=0;i<PDI_LAYERS;i++) vlb_vid_in[i] = '{VCLK_PHASE_IN,H_ena_in,V_ena_in,VENA_in,HS_in,VS_in,WLENA_in[i],RGBA_in[i]};

// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// CMD_win_alpha_adj for each layer.
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
generate if (!ENABLE_alpha_adj) begin
    assign alpha_out = vlb_vid_in ;
end else begin

logic        [7:0] alpha_adj    [0:PDI_LAYERS-1] ;
logic        [8:0] alpha_fixp   [0:PDI_LAYERS-1] ;
logic        [8:0] alpha_fixn   [0:PDI_LAYERS-1] ;

always_comb begin
    for (int i=0 ; i<PDI_LAYERS; i++) begin
    
        alpha_fixp[i] = (vlb_vid_in[i].rgba[7:0] +           {alpha_adj_in[i][6:0],alpha_adj_in[i][6]}  ) ;
        alpha_fixn[i] = (vlb_vid_in[i].rgba[7:0] - 8'(8'd255^{alpha_adj_in[i][6:0],alpha_adj_in[i][6]}) ) ;

        // Use the active window area flag WLENA_in[] to ensure a transparent alpha no matter the alpha adjust setting for that layer.
        if (!alpha_adj_in[i][7]) alpha_adj[i] = !WLENA_in[i] ? 8'd0 : alpha_fixp[i][8] ? 8'd255 : alpha_fixp[i][7:0] ; // Add, so overshoot =255
        else                     alpha_adj[i] = !WLENA_in[i] ? 8'd0 : alpha_fixn[i][8] ? 8'd0   : alpha_fixn[i][7:0] ; // subtract, so overshoot =0


    end
end

    always_ff @(posedge VID_CLK) begin

    for (int i=0 ; i<PDI_LAYERS; i++) begin
                                        alpha_out[i].rgba  <= {vlb_vid_in[i].rgba[31:8],alpha_adj[i]};
                                        alpha_out[i].wlena <=  vlb_vid_in[i].wlena ;
                                        alpha_out[i].phase <=  vlb_vid_in[i].phase ;
                                        alpha_out[i].h_ena <=  vlb_vid_in[i].h_ena ;
                                        alpha_out[i].v_ena <=  vlb_vid_in[i].v_ena ;
                                        alpha_out[i].vena  <=  vlb_vid_in[i].vena  ;
                                        alpha_out[i].hs    <=  vlb_vid_in[i].hs    ;
                                        alpha_out[i].vs    <=  vlb_vid_in[i].vs    ;
                                        end
    end
end endgenerate

// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// SDI sequence phase swapper on each PDI layer.
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
generate if (!ENABLE_SDI_layer_swap || SDI_LAYERS==1) begin
    assign sdi_swap = alpha_out;
end else begin
    assign sdi_swap = alpha_out;
end endgenerate


// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// PDI channel swapper at each SDI phase position.
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
generate if (!ENABLE_PDI_layer_swap || PDI_LAYERS==1) begin
    assign pdi_swap = sdi_swap;
end else begin
    assign pdi_swap = sdi_swap;
end endgenerate



// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// Progressive PDI layer mixer.
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
generate if (PDI_LAYERS==1) begin
    assign pdi_mix = pdi_swap[0];
end else begin

    PDI_mixer #(.in_layers(2)) mix_pdi (.clk(VID_CLK),.in(pdi_swap[0:1]),.out(pdi_mix_out));

    assign pdi_mix = pdi_mix_out[0];
end endgenerate


// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// Final sequential SDI layer mixer.
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
generate if (SDI_LAYERS==1) begin
    assign sdi_mix = pdi_mix ;
end else begin

    // When mixing the initial bottom layer, assign the BGC background color to the layer beneath that one.
    // Otherwise, use the previous calculated mix layer as the new bottom layer.
    logic mix_bgc ;
    wire [7:0] mix_bot_r = (mix_bgc || (SDI_LAYERS==1)) ? CMD_BGC_RGB[23:16] : sdi_mix.rgba[31:24] ;
    wire [7:0] mix_bot_g = (mix_bgc || (SDI_LAYERS==1)) ? CMD_BGC_RGB[15: 8] : sdi_mix.rgba[23:16] ;
    wire [7:0] mix_bot_b = (mix_bgc || (SDI_LAYERS==1)) ? CMD_BGC_RGB[ 7: 0] : sdi_mix.rgba[15: 8] ;
    wire [7:0] mix_r ;
    wire [7:0] mix_g ;
    wire [7:0] mix_b ;

    // Calculate the mix using the previous layer as the bottom and the current pixel data as the top layer RGB and alpha in the mixing process.
    mixer #(.mbits(8),.dbits(8)) mix_sdi_r (.mix(pdi_mix.rgba[7:0]),.bot(mix_bot_r),.top(pdi_mix.rgba[31:24]),.out(mix_r));
    mixer #(.mbits(8),.dbits(8)) mix_sdi_g (.mix(pdi_mix.rgba[7:0]),.bot(mix_bot_g),.top(pdi_mix.rgba[23:16]),.out(mix_g));
    mixer #(.mbits(8),.dbits(8)) mix_sdi_b (.mix(pdi_mix.rgba[7:0]),.bot(mix_bot_b),.top(pdi_mix.rgba[15: 8]),.out(mix_b));

    // Sequentially clock the pixel pipe.
    always_ff @(posedge VID_CLK) begin

    mix_bgc <= (pdi_mix.phase==0) ; // Prepare the next cycle to have the bottom BGC mix color.

    sdi_mix.rgba  <= {mix_r,mix_g,mix_b,pdi_mix.rgba[7:0]};
    sdi_mix.wlena <= pdi_mix.wlena ;
    sdi_mix.phase <= pdi_mix.phase ;
    sdi_mix.h_ena <= pdi_mix.h_ena ;
    sdi_mix.v_ena <= pdi_mix.v_ena ;
    sdi_mix.vena  <= pdi_mix.vena  ;
    sdi_mix.hs    <= pdi_mix.hs    ;
    sdi_mix.vs    <= pdi_mix.vs    ;

    end

end endgenerate

// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// Assign outputs.
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
assign RGBA_out       = sdi_mix.rgba  ;
assign VCLK_PHASE_OUT = sdi_mix.phase ;
assign H_ena_out      = sdi_mix.h_ena ;
assign V_ena_out      = sdi_mix.v_ena ;
assign VENA_out       = sdi_mix.vena  ;
assign HS_out         = sdi_mix.hs    ;
assign VS_out         = sdi_mix.vs    ;

endmodule


// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// Basic TOP/BOTTOM alpha blending mixer.
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
module mixer #(parameter mbits=8, parameter dbits=8)(input [mbits-1:0] mix, input [dbits-1:0] top, input [dbits-1:0] bot, output [dbits-1:0] out);

wire [mbits-1:0] mtop = mix;
wire [mbits-1:0] mbot = ((mbits)'(2**mbits-1))-mix;

// The added LSB bit set to 1 in the multipliers helps retain full contrast without loosing a single shade.
// This is necessary since running the display through multiple layers would otherwise mean loosing 1 brightness per mixing layer.
// IE: 16 layers means a big loss of contrast, 16 out of 255.

wire [mbits+dbits+1:0] fact_top = {mtop,1'b1} * {top,1'b1} ; 
wire [mbits+dbits+1:0] fact_bot = {mbot,1'b1} * {bot,1'b1} ; 

assign out = (dbits)'((fact_top>>(mbits+2))+(fact_bot>>(mbits+2)));

endmodule

// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
// Mix the bottom 2 layers together and pass through the remaining layers,
// **********************************************************************************
// **********************************************************************************
// **********************************************************************************
module PDI_mixer #(parameter in_layers=2)(input clk, input vlb_bus in[0:in_layers-1], output vlb_bus out [0:in_layers-2]);

    wire [7:0] mix_r ;
    wire [7:0] mix_g ;
    wire [7:0] mix_b ;

    mixer #(.mbits(8),.dbits(8)) mix_pdi_r (.mix(in[in_layers-2].rgba[7:0]),.bot(in[in_layers-1].rgba[31:24]),.top(in[in_layers-2].rgba[31:24]),.out(mix_r));
    mixer #(.mbits(8),.dbits(8)) mix_pdi_g (.mix(in[in_layers-2].rgba[7:0]),.bot(in[in_layers-1].rgba[23:16]),.top(in[in_layers-2].rgba[23:16]),.out(mix_g));
    mixer #(.mbits(8),.dbits(8)) mix_pdi_b (.mix(in[in_layers-2].rgba[7:0]),.bot(in[in_layers-1].rgba[15: 8]),.top(in[in_layers-2].rgba[15: 8]),.out(mix_b));


    always_ff @(posedge clk) begin

    out[in_layers-2].rgba  <= {mix_r,mix_g,mix_b,in[in_layers-2].rgba[7:0]};
    out[in_layers-2].wlena <= in[in_layers-2].wlena ;
    out[in_layers-2].phase <= in[in_layers-2].phase ;
    out[in_layers-2].h_ena <= in[in_layers-2].h_ena ;
    out[in_layers-2].v_ena <= in[in_layers-2].v_ena ;
    out[in_layers-2].vena  <= in[in_layers-2].vena  ;
    out[in_layers-2].hs    <= in[in_layers-2].hs    ;
    out[in_layers-2].vs    <= in[in_layers-2].vs    ;

        //for (int i=in_layers-2 ; i>0 ; i--) out[i-1] <= in[i] ; // Push up and pass through the remaining layers. 

    end

endmodule
